/*
MAXFLOW MAIN-METHOD by ron

  0 - 1 - 2             0 - 1 - 2
  | +-+-+ |	        |   |   |
  3 + 4 + 5     <====   3 - 4 - 5
--+-+ | +-+--           |   |   |
  6 - 7 - 8             6 - 7 - 8
cut trough G            graph G

ich habe diesen graphen unten initialisiert. die kantengewichte habe ich dabei auf 1 gelassen.
wie man sehen kann ist der graph als ungerichtet zu betrachten. der versuchte schnitt sollte die pixel 0,1,2,3,5 extrahieren.
unter diesen bedingungen war der schnitt erfolgreich. waehlte man allerdings die kanten gewichte asymmetrisch,so kann er nicht
immer saemtliche pixel extrahieren.

resultate (may 06):
die resultate koennt ihr gerne selber einmal ausprobieren. mir hat es einige einblicke in die wahl der gewichte gebracht.
wie muessen diese gewichte gewaehlt werden? mit ausblick darauf, dass wir mal ein bild in unsere applikation eingeben wollen,
kann man wohl sagen, dass die gewichte anhand der farben ausgerechnet werden. dabei ist jedes pixel ein knoten im graphen.
so wie ich mir das vorstelle, werden die gewichte anhand der farbunterschiede ausgerechnet, was jedoch wieder einen
ungerichteten graphen ergibt. vielleicht laesst sich eine light-version des codes basteln, der rein mit ungerichteten graphen
arbeitet und etwas schneller dann laeuft.

ausblick:
so wie ich mir das vorstelle, gibt der user ein bild ein. dann waehlt er grob den vordergrund und hintergrund. danach baut
der algorithmus sich den graphen zusammen. dazu gehoert:
	.: nodes (=pixel)
	.: edges (vorgegeben, da bitmap - es muessen lediglich die kantengewichte berechnet werden)
	.: labels (anhand von maxflow.cpp)

update (jun 06):
endlich konnte die funktion in ein framework einbaut werden und funktioniert jetzt als plugin. 


andre schmidt 06
*/

#include <stdlib.h>
#include <iostream.h>
#include <math.h>

#ifndef M_PI
#  define  M_PI  3.14159265358979323846     // in case pi wasn't defined in math.h
#endif

// Header for the application framework:
#include <framewk.h>
#include "graph.h"

const int X_SIZE = 512; //bitmap in x richtung
const int Y_SIZE = 512; //bitmap in y richtung
Graph::node_id nodes[262144];

char graphics[X_SIZE/8][Y_SIZE]; //.../8 da 8 bits pro byte


void graphCut()
{

	/* 
	*  get information from the original image
	*  the chosen picture format is TIFF:
	*/
	ImageRGBA*  img=  imageRgbaReadTIFF("in.tif");
	int pic_w = img->width;
	int pic_h = img->height;
	int pic_size = pic_h*pic_w;


	/* 
	*  get information of the user chosen back and foreground
	*  the chosen picture format is TIFF:
	*/
	ImageRGBA*  img_sel=  imageRgbaReadTIFF("sel.tif");

	//build graph
	Graph *g = new Graph();
	for(int i=0;i<pic_size;i++){
		nodes[i] = g -> add_node();
	}

	//get source/sink information
	Rgba rgb;
	Rgba picrgb;
	int valR, valG;
	
	int meanvalRsink = 0;
	int meanvalGsink = 0;
	int meanvalBsink = 0;
	int meanvalRsource = 0;
	int meanvalGsource = 0;
	int meanvalBsource = 0;
	int sinkNodes = 0;
	int sourceNodes = 0;
	
	for(int row=0;row<pic_h;row++){
		for(int col=0;col<pic_w;col++){
			rgb = img_sel->data[col+pic_w*row];
			valR = (int)rgb.r;
			valG = (int)rgb.g;
			if(valR==255 && valG==0){
				g -> set_tweights(nodes[row*pic_w+col], 0, 766);

				sinkNodes++;
				picrgb = img->data[col+pic_w*row];
				meanvalRsink += picrgb.r;
				meanvalGsink += picrgb.g;
				meanvalBsink += picrgb.b;
			}
			if(valR==0 && valG==255){
				g -> set_tweights(nodes[row*pic_w+col], 766, 0);

				sourceNodes++;
				picrgb = img->data[col+pic_w*row];
				meanvalRsource += picrgb.r;
				meanvalGsource += picrgb.g;
				meanvalBsource += picrgb.b;
			}
		}
	}

	//mean color of sink nodes
	meanvalRsink = meanvalRsink/sinkNodes;
	meanvalGsink = meanvalGsink/sinkNodes;
	meanvalBsink = meanvalBsink/sinkNodes;

	//mean color of source nodes
	meanvalRsource = meanvalRsource/sourceNodes;
	meanvalGsource = meanvalGsource/sourceNodes;
	meanvalBsource = meanvalBsource/sourceNodes;


	Rgba rgbP;
	int valRP, valGP, valBP;
	int diffASink;
	int diffASource;
	for(int row=0;row<pic_h;row++){
		for(int col=0;col<pic_w;col++){
			rgbP = img->data[col+pic_w*row];
			valRP = (int)rgbP.r;
			valGP = (int)rgbP.g;
			valBP = (int)rgbP.b;
			
			//differenz zwischen farbe des aktuellen punktes und der mittelwerte von source und sink berechnen
			//wir haben uns am ende für die l2-norm des farbvektors entschieden, da es bessere resultate liefert.
			//diffASink = abs(valRP - meanvalRsink) + abs(valGP - meanvalGsink) + abs(valBP - meanvalGsink);
			//diffASource = abs(valRP - meanvalRsource) + abs(valGP - meanvalGsource) + abs(valBP - meanvalGsource);
			diffASink = sqrt(pow(valRP - meanvalRsink,2) + pow(valGP - meanvalGsink,2) + pow(valBP - meanvalGsink,2));
			diffASource = sqrt(pow(valRP - meanvalRsource,2) + pow(valGP - meanvalGsource,2) + pow(valBP - meanvalGsource,2));


			//falls farbe des punktes ähnlich wie source oder sink -> tweights setzen!
			if (abs((diffASink-diffASource) > 50) && !(g -> tweight_set(nodes[row*pic_w+col])))  {
				if (diffASink < 50) {
					g -> set_tweights(nodes[row*pic_w+col], 0, 200);
				}

				if (diffASource < 50) {
					g -> set_tweights(nodes[row*pic_w+col], 200, 0);
				}
			}			
		}
	}


	//make connection between nodes (edges)
	Rgba rgbA, rgbB;
	int valRA, valRB, valGA, valGB, valBA, valBB;
	int diffAB;
	for(int row=1;row<pic_h;row++){
		for(int col=1;col<pic_w;col++){
			rgbA = img->data[col+pic_w*row];
			rgbB = img->data[col-1+pic_w*(row-1)];
			valRA = (int)rgbA.r;
			valGA = (int)rgbA.g;
			valBA = (int)rgbA.b;
			valRB = (int)rgbB.r;
			valGB = (int)rgbB.g;
			valBB = (int)rgbB.b;

			//gewichtung der kanten: abstand des farbvektors im rbg-raum.
			diffAB = sqrt(pow(valRA - valRB,2) + pow(valGA - valGB,2)+ pow(valBA - valBB,2));

			//gewichtung der kanten: abstand der farbwerte addiert.
			//diffAB = abs(valRA - valRB) + abs(valGA - valGB) + abs(valBA - valBB);

			
			if(diffAB==0)
				diffAB = 1;

			g -> add_edge(nodes[row*pic_w+col], nodes[row*pic_w+col-1], 765/diffAB, 765/diffAB);
			g -> add_edge(nodes[row*pic_w+col], nodes[(row-1)*pic_w+col], 765/diffAB, 765/diffAB);
		}
	}

	Graph::flowtype flow = g -> maxflow();

	//give result of maxflow
	printf("Flow = %d\n", flow);
	for(int row=0;row<pic_h;row++){
		for(int col=0;col<pic_w;col++){
			if (g->what_segment(nodes[row*pic_w+col]) == Graph::SINK){
				Rgba whiteDot;
				whiteDot.r = 255;
				whiteDot.g = 255;
				whiteDot.b = 255;
				img->data[col+pic_w*row] = whiteDot;
			}
		}
	}

	//display and end
	delete g;
	imageRgbaWriteTIFF("res.tif", img);

}

