//
//    Title: painter.cc
//  Created: Tue Oct  2 15:53:57 2001
//   Author: Tim Weyrich <weyrich@inf.ethz.ch>
//
//  copyright (c) 2001--2003, Computer Graphics Laboratory, ETH Zuerich
//

#include <stdlib.h>
#include <iostream.h>
#include <math.h>

#ifndef M_PI
#  define  M_PI  3.14159265358979323846     // in case pi wasn't defined in math.h
#endif

// Header for the application framework:
#include <framewk.h>

//Header for the graphCut method:
#include "ais.h"

//
//  See http://graphics.ethz.ch/~weyrich/Info-I/painter/ for a
//  program documentation...
//

ImageRGBA  *myImage = NULL;
ImageRGBA  *orgImage = NULL;

bool   mousePressed = false;
float  mouseX = -1.0f;
float  mouseY = -1.0f;
bool   mousePressedR = false;
char   mode = ' ';
int    cut = 0;
int    brush = 5;
const char* fileName;


///////////////////////////////////////////////////////////////////////////////
// A Sample Effect Called `Hoover'                                           //
///////////////////////////////////////////////////////////////////////////////

void  hoover(ImageRGBA *img, float mouseX, float mouseY)
  //! Applies my `hoover effect' to an image.
  /*! This function implements a nice effect on an image. It is not
   *  necessary, that you understand this code in total. But for the
   *  interested, it may show, that even more spectacular actions on
   *  images may be coded easily...
   */
{
  // Hints to understand the code:
  //
  //   rand() returns an integer out of 0..RAND_MAX.RAND__MAX is a
  //   very large integer.
  //
  // 
  //   Using rand(),
  const int   w = img->width;
  const int   h = img->height;

  if (img!=NULL && mousePressed)
    {
      for (int i=50000; i; i--)
        {
          double  phi = 2.0*3.14159*(rand() & 32767)/32768;
          double  r = rand() % w;
          
          double  c = cos(phi);
          double  s = sin(phi);

          int  x1 = (int)(mouseX + r * c);
          int  y1 = (int)(mouseY + r * s);

          if (x1 >= 0 && x1 < w && y1 >= 0 && y1 < h)
            {
              int  x2 = (int)(mouseX + (r-1.0) * c);
              int  y2 = (int)(mouseY + (r-1.0) * s);
              
              img->data[x2 + y2*w] = img->data[x1 + y1*w];
            }
        }
    }
}


///////////////////////////////////////////////////////////////////////////////
// Draw a thick line into the pic to select background                       //
///////////////////////////////////////////////////////////////////////////////

void  markRed(ImageRGBA *img, float mouseX, float mouseY)
  /*! This function draws a thick line into the picture.
   */
{
 	Rgba colorValue;
	colorValue.r=255;
	colorValue.b=0;
	colorValue.g=0;
	colorValue.a=0;
	imageRgbaDrawEllipse(img, mouseX, mouseY, brush, brush, colorValue);

}


///////////////////////////////////////////////////////////////////////////////
// Draw a thick line into the pic to select foreground                       //
///////////////////////////////////////////////////////////////////////////////

void  markGreen(ImageRGBA *img, float mouseX, float mouseY)
  /*! This function draws a thick line into the picture.
   */
{
 	Rgba colorValue;
	colorValue.r=0;
	colorValue.b=0;
	colorValue.g=255;
	colorValue.a=0;
	imageRgbaDrawEllipse(img, mouseX, mouseY, brush, brush, colorValue);

}


///////////////////////////////////////////////////////////////////////////////
// Restore the original image with the right mouse button                    //
///////////////////////////////////////////////////////////////////////////////
 
void  restoreImg(ImageRGBA *img, float mouseX, float mouseY)
{
	const int   w = img->width;
	if(!orgImage){
		orgImage = imageRgbaReadTIFF(fileName);
	}
      	for(int x=((int)mouseX)-3;x < ((int)mouseX)+4;x++){
		for(int y=((int)mouseY)-3;y < ((int)mouseY)+4;y++){
      			img->data[x + w * y] = orgImage->data[x + w * y];
      		}
      	}
}


///////////////////////////////////////////////////////////////////////////////
// Invert selection on the picture                                           //
///////////////////////////////////////////////////////////////////////////////

void  invertSelection(ImageRGBA *img)
{
	const int   w = img->width;
	const int   h = img->height;
	if(!orgImage){
		orgImage = imageRgbaReadTIFF(fileName);
	}
      	for(int x=0;x<w;x++){
		for(int y=0;y<h;y++){
			if(img->data[x + w * y].r == 255 && img->data[x + w * y].g == 0 && img->data[x + w * y].b == 0){
      				img->data[x + w * y].r = 0;
				img->data[x + w * y].g = 255;
			}
			else{
				if(img->data[x + w * y].g == 255 && img->data[x + w * y].r == 0 && img->data[x + w * y].b == 0){
      					img->data[x + w * y].r = 255;
					img->data[x + w * y].g = 0;
				}
			}
      		}
      	}
}



//////////////////////////////////////////////////////////////////////////////////
// Display the rgba picture with only gray colors (leaving the image in rgba)   //
//////////////////////////////////////////////////////////////////////////////////

void  makeGray(ImageRGBA *img)
  /*! This function replaces all colors by gray tones.
   */
{
	Rgba cval;
	int val;

 	for (int mouseY=0;mouseY < img->height; mouseY++){
		for (int mouseX=0;mouseX < img->width; mouseX++){
			cval = img->data[mouseY*img->width+mouseX];
			val = ((double)cval.r) * 0.3 + ((double)cval.g) * 0.59 + ((double)cval.b) * 0.11;
			cval.r = val;
			cval.g = val;
			cval.b = val;
			img->data[mouseY*img->width+mouseX] = cval;
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
// Function for calling external bin                                         //
///////////////////////////////////////////////////////////////////////////////

/*
"prg" is the binary to execute
"arg1" is an argument1 for the binary
"arg2" is an argument2 for the binary
"type" controls the output: 1 is stdout, 2 is stderr
	The funtion returns:
	-1 in case of an incorrect "type"
	-2 if no pipe could be created
	-3 if a process couldn't be forked
"str" a string from stdout or stderr.
*/

char* execSub(char *prg, char *arg1, char *arg2, int type)
{
	pid_t Pid;
	int pipefd[2];
	char buffer[5000];
	int anz, i, status=0;
	char *str;

	status=0;
	anz=0;
	str=NULL;

	/* pipe is | */
	if(pipe(pipefd)==-1){ //PIPE-error
		str=(char*) malloc(2);
		strcpy(str, "-2");
		return str;}

	switch((Pid=fork())) {
		case 0:{ //---------------Child
			close(pipefd[0]);
			if(type==0)
				close(1);	
			if(type==1)
				close(2);
			else{
				str=(char*) malloc(2);
				strcpy(str, "-1");
			return str;}
			dup(pipefd[1]);
			execlp(prg, prg, arg1, arg2, NULL); //prg gets executed
			exit(0);
		}
		case -1: { //FORK-error
			str=(char*) malloc(2);
			strcpy(str, "-3");
			return str;}
		default:{ //---------------Parent
			//wait(&status);
			close(pipefd[1]);
			printf(" executing %s %s %s ", prg, arg1, arg2);
			while((anz=read(pipefd[0],buffer,5000))>0){
			if(str==NULL)
				str=(char*) calloc((anz+1), sizeof(anz));
			else if(buffer==NULL && str !=NULL)
				str=(char*)realloc(str, sizeof(anz+1));
				for(i=0; i<anz;i++)
					str[i]=buffer[i];
			}
		}
	return str;
	}
}



///////////////////////////////////////////////////////////////////////////////
// Callback Functions                                                        //
///////////////////////////////////////////////////////////////////////////////

void  keyboardCB(unsigned char key, float x, float y)
  /*! This callback function is called, when a non-special key (see
   *  specialCB()) was pressed.
   *
   *   \param key    The ASCII code of the key, that was pressed.
   *
   *   \param x, y   Mouse position, when the key was pressed.
   *                 The position is measured in image coordinates.
   *                 Note that x and y can be fractional numbers if
   *                 the image is zoomed.  If the mouse is outside the
   *                 window, it may happen that x and y are outside
   *                 the image boundaries.
   */
{
  (void)x; (void)y; // This line prevents the compiler warning `variables not used'.
  
  switch (key)
    {
    case 's':
      imageRgbaWriteTIFF("out.tif", myImage);
      break;
      
    case '+':
    case '=':
      fwkSetDisplayZoom(1.25 * fwkGetDisplayZoom());
      break;
      
    case '-':
      fwkSetDisplayZoom(0.8 * fwkGetDisplayZoom());
      break;
      
    case 'q':
      exit(0);
      break;

    case 'o':
      cut = 0;
      myImage = imageRgbaReadTIFF(fileName);
      orgImage = imageRgbaReadTIFF(fileName);;
      fwkSetDisplayImage(myImage);
      fwkPostRedisplay();
      break;

    case 'g':
      makeGray(myImage);
      makeGray(orgImage);
      fwkSetDisplayImage(myImage);
      fwkPostRedisplay();
      break;

    case 'h':
      mode = 'h';
      break;

    case 'f':
      if(cut >=1  && cut < 3 && mode != 'f')
        cut++;
      mode = 'f';
      break;

    case 'b':
      if(cut >=1 && cut < 3 && mode != 'b')
        cut++;
      mode = 'b';
      break;

    case 'c':
      cout << "Press 'f' to select foreground. Afterwards press 'b' to select background.\nWhen finished press 'enter' to perform the cut." << endl;
      cut = 1;
      mode = ' ';
      makeGray(myImage);
      makeGray(orgImage);
      fwkSetDisplayImage(myImage);
      fwkPostRedisplay();
      break;

     case 13:
      if(cut == 3){
	cut = 4;
        imageRgbaWriteTIFF("sel.tif", myImage);
        graphCut();
        myImage = imageRgbaReadTIFF("res.tif");
        orgImage = imageRgbaReadTIFF("res.tif");
        fwkSetDisplayImage(myImage);
        fwkPostRedisplay();
      }
      else{
        cout <<'\n' <<endl;
      }
      break;

    case 'C':
      if(cut==4){
	cut = 3;
	cout << "Press 'f' to select foreground. Afterwards press 'b' to select background.\nWhen finished press 'enter' to perform the cut." << endl;
        mode = ' ';
        myImage = imageRgbaReadTIFF("sel.tif");
        orgImage = imageRgbaReadTIFF(fileName);
	makeGray(orgImage);
        fwkSetDisplayImage(myImage);
        fwkPostRedisplay();
      }
      else{
        cout << "Press 'c' to define cut first." << endl;
      }
      break;

    case 'i':
      if(cut==3){
	cout << "Press 'f' to select foreground. Afterwards press 'b' to select background.\nWhen finished press 'enter' to perform the cut." << endl;
        mode = ' ';
        myImage = imageRgbaReadTIFF("sel.tif");
	invertSelection(myImage);
        orgImage = imageRgbaReadTIFF(fileName);
	makeGray(orgImage);
        fwkSetDisplayImage(myImage);
        fwkPostRedisplay();
      }
      else{
        cout << "Press 'c' to define cut first." << endl;
      }
      break;

    case '?':
      cout
        << "Mouse Shortcuts:" << endl
        << "-------------------------------------------------------------------------------" << endl
        << "  Left Mousebutton      Draw on the picture." << endl
	<< "  Right Mousebutton     Erase user paint on the picture." << endl
        << "\n\n" << endl
        << "Keyboard Shortcuts:" << endl
        << "-------------------------------------------------------------------------------" << endl
        << "  s                     Save image as `out.tif'." << endl
	<< "  o                     Restore original image." << endl
        << "  h                     Perform hoover effect." << endl
	<< "  f                     Select foreground for GraphCut." << endl
	<< "  b                     Select background for GraphCut." << endl
	<< "  c                     Lauch kick ass GraphCut algorithm." << endl
	<< "  C                     Jump back to last selection." << endl
	<< "  i                     Invert selection." << endl
	<< "  F1-F10                Change brush size." << endl
        << "  +/-                   Zoom in/out." << endl
        << "  q                     Quit this application." << endl
        << "  ?                     This help." << endl;
      break;
      
    default:
      cout << "\aKey `" << key << "' (code " << (int)key << ") is not defined." << endl;
    }
}

void  specialCB(int key, float x, float y)
  /*! This callback function is called, when a `special key', i.e. a
   *  key with no printable representation, was pressed.
   *
   *   \param key    The special code of the key, that was pressed.
   *                 Symbolic constants for the possible values of
   *                 key can be found in GL/glut.h. (On the Sun
   *                 computers in the exercise pool, this file is
   *                 located at
   *                 /usr/pack/glut-3.7-mo/solaris/include/GL/glut.h.)
   *
   *   \param x, y   Mouse position, when the key was pressed.
   *                 The position is measured in image coordinates.
   *                 Note that x and y can be fractional numbers if
   *                 the image is zoomed.  If the mouse is outside the
   *                 window, it may happen that x and y are outside
   *                 the image boundaries.
   */
{
  (void)x; (void)y; // This line prevents the compiler warning `variables not used'.

  switch(key)
    {

    case GLUT_KEY_F1:
      // choose brush thickness
      brush = 0;
      cout << "Brush thickness 0." << endl;
      break;

    case GLUT_KEY_F2:
      // choose brush thickness
      brush = 1;
      cout << "Brush thickness 1." << endl;
      break;

    case GLUT_KEY_F3:
      // choose brush thickness
      brush = 2;
      cout << "Brush thickness 2." << endl;
      break;

    case GLUT_KEY_F4:
      // choose brush thickness
      brush = 3;
      cout << "Brush thickness 3." << endl;
      break;

    case GLUT_KEY_F5:
      // choose brush thickness
      brush = 4;
      cout << "Brush thickness 4." << endl;
      break;

    case GLUT_KEY_F6:
      // choose brush thickness
      brush = 5;
      cout << "Brush thickness 5." << endl;
      break;

    case GLUT_KEY_F7:
      // choose brush thickness
      brush = 6;
      cout << "Brush thickness 6." << endl;
      break;

    case GLUT_KEY_F8:
      // choose brush thickness
      brush = 7;
      cout << "Brush thickness 7." << endl;
      break;

    case GLUT_KEY_F9:
      // choose brush thickness
      brush = 8;
      cout << "Brush thickness 8." << endl;
      break;

    case GLUT_KEY_F10:
      // choose brush thickness
      brush = 9;
      cout << "Brush thickness 9." << endl;
      break;


    default:
      cout << "\aCommand not defined." << endl;
    }
}

void  mouseCB(int button, int state, float x, float y)
  /*! This callback function is called, when a mouse button was
   *  pressed or released.
   *
   *   \param button         The number of the button, that was
   *                         pressed. \a button can be one of
   *                         the following values: GLUT_LEFT_BUTTON,
   *                         GLUT_MIDDLE_BUTTON, GLUT_RIGHT_BUTTON.
   *
   *   \param state          Denotes, whether the button was
   *                         pressed or released. \a state can be
   *                         be one of the following values:
   *                         GLUT_DOWN, when the button was pressed,
   *                         or GLUT_UP, when the button was
   *                         released.
   *
   *   \param x, y           Mouse position, when the button was
   *                         pressed/released.
   *                         The position is measured in image
   *                         coordinates. Note that x and y can be
   *                         fractional numbers if the image is
   *                         zoomed.  If the mouse is outside the
   *                         window, it may happen that x and y are
   *                         outside the image boundaries.
   */
{
 
  if (button == GLUT_LEFT_BUTTON)
    if (state == GLUT_DOWN)
      {
        //cout << "Left mouse button pressed." << endl;

        // Save current image coordinates in mouseX/mouseY.
        mouseX = x;
        mouseY = y;
        mousePressed = true;
      }
    else
      {
        mousePressed = false;
      }

  if (button == GLUT_RIGHT_BUTTON)
    if (state == GLUT_DOWN)
      {
        //cout << "Left mouse button pressed." << endl;

        // Save current image coordinates in mouseX/mouseY.
        mouseX = x;
        mouseY = y;
        mousePressedR = true;
      }
    else
      {
        mousePressedR = false;
      }
}

void  motionCB(float x, float y)
  /*! This callback function is called, when the mouse is moved,
   *  while a mouse button is held pressed.
   *
   *   \param x, y           Current mouse position.
   *                         The position is measured in image
   *                         coordinates. Note that x and y can be
   *                         fractional numbers if the image is
   *                         zoomed.  If the mouse is outside the
   *                         window, it may happen that x and y are
   *                         outside the image boundaries.
   */
{
  // Here is an example code, that shows, how to react on mouse
  // movements.
  //
  // The image coordinates are stored in mouseX and mouseY, so you can
  // read out this variables in other call-back functions.  Note, that
  // mouseX and mouseY are only valid, if mousePressed is `true'!
  //
  {
    mouseX = x;
    mouseY = y;
    
    /*coutpainter.cpp:522:
      << "motion event : mouse pointer at ("
      << mouseX << ", " << mouseY <<
      ") in image coordinates." << endl;  */
  }
  
}

void  idleCB(void)
  /*! This callback function is called, whenever the application
   *  main-loop is idle.
   */
{
  // To present an example use of the idleCB-Function, we call our
  // `hoover effect', when the application main-loop is idle and the
  // left mouse button is pressed...
  if (myImage && mousePressed && mode=='f')
    {
      markGreen(myImage, mouseX, mouseY);
      fwkPostRedisplay(); // Don't forget this! If we don't call
      //                     fwkPostRedisplay, we'll see nothing
      //                     on the screen...
    }

  if (myImage && mousePressed && mode=='b')
    {
      markRed(myImage, mouseX, mouseY);
      fwkPostRedisplay(); // Don't forget this! If we don't call
      //                     fwkPostRedisplay, we'll see nothing
      //                     on the screen...
    }  

  if (myImage && mousePressed && mode=='h')
    {
      hoover(myImage, mouseX, mouseY);
      fwkPostRedisplay(); // Don't forget this! If we don't call
      //                     fwkPostRedisplay, we'll see nothing
      //                     on the screen...
    }  

  if (myImage && mousePressedR)
    {
     	restoreImg(myImage, mouseX, mouseY);
	fwkPostRedisplay();
     }

}


///////////////////////////////////////////////////////////////////////////////
// main                                                                      //
///////////////////////////////////////////////////////////////////////////////

int  main(int argc, const char *argv[])
  /*! See \ref pgdocFramework for information, how to write your
   *  main() function.
   */
{
  // Important: Initialize application framework.
  fwkInit(&argc, argv);

  {
    fileName = argv[1];	
    if(!fileName){
	cout << "Please define input file as first parameter!." << endl;
	return 1;
    }
    myImage = imageRgbaReadTIFF(argv[1]);
    orgImage = imageRgbaReadTIFF(argv[1]);
    fwkSetDisplayImage(myImage);
    imageRgbaWriteTIFF("in.tif", myImage);
  }

  // Start application main-loop.
  fwkMainLoop();
  
  return 1; // Return an error value, since we should never reach this point!
}

///////////////////////////////////////////////////////////////////////////////
